package java.lang.reflect;

import sun.reflect.annotation.TypeAnnotationParser;
import sun.reflect.annotation.TypeAnnotation;
import sun.misc.SharedSecrets;
import sun.reflect.annotation.AnnotationParser;
import sun.reflect.annotation.AnnotationSupport;
import java.util.Objects;
import sun.reflect.CallerSensitive;
import sun.reflect.Reflection;
import sun.reflect.generics.factory.CoreReflectionFactory;
import sun.reflect.generics.scope.ClassScope;
import sun.reflect.generics.factory.GenericsFactory;
import java.lang.annotation.Annotation;
import java.util.Map;
import sun.reflect.FieldAccessor;
import sun.reflect.generics.repository.FieldRepository;

public final class Field extends AccessibleObject implements Member
{
    private Class<?> clazz;
    private int slot;
    private String name;
    private Class<?> type;
    private int modifiers;
    private transient String signature;
    private transient FieldRepository genericInfo;
    private byte[] annotations;
    private FieldAccessor fieldAccessor;
    private FieldAccessor overrideFieldAccessor;
    private Field root;
    private transient Map<Class<? extends Annotation>, Annotation> declaredAnnotations;
    
    private String getGenericSignature() {
        return this.signature;
    }
    
    private GenericsFactory getFactory() {
        Class<Object> declaringClass = this.getDeclaringClass();
        return CoreReflectionFactory.make(declaringClass, ClassScope.make(declaringClass));
    }
    
    private FieldRepository getGenericInfo() {
        if (this.genericInfo == null) {
            this.genericInfo = FieldRepository.make(this.getGenericSignature(), this.getFactory());
        }
        return this.genericInfo;
    }
    
    Field(Class<?> clazz, String name, Class<?> type, int modifiers, int slot, String signature, byte[] annotations) {
        this.clazz = clazz;
        this.name = name;
        this.type = type;
        this.modifiers = modifiers;
        this.slot = slot;
        this.signature = signature;
        this.annotations = annotations;
    }
    
    Field copy() {
        if (this.root != null) {
            throw new IllegalArgumentException("Can not copy a non-root Field");
        }
        Field field = new Field(this.clazz, this.name, this.type, this.modifiers, this.slot, this.signature, this.annotations);
        field.root = this;
        field.fieldAccessor = this.fieldAccessor;
        field.overrideFieldAccessor = this.overrideFieldAccessor;
        return field;
    }
    
    @Override
    public Class<?> getDeclaringClass() {
        return this.clazz;
    }
    
    @Override
    public String getName() {
        return this.name;
    }
    
    @Override
    public int getModifiers() {
        return this.modifiers;
    }
    
    public boolean isEnumConstant() {
        return (this.getModifiers() & 0x4000) != 0x0;
    }
    
    @Override
    public boolean isSynthetic() {
        return Modifier.isSynthetic(this.getModifiers());
    }
    
    public Class<?> getType() {
        return this.type;
    }
    
    public Type getGenericType() {
        if (this.getGenericSignature() != null) {
            return this.getGenericInfo().getGenericType();
        }
        return this.getType();
    }
    
    @Override
    public boolean equals(Object o) {
        if (o != null && o instanceof Field) {
            Field field = (Field)o;
            return this.getDeclaringClass() == field.getDeclaringClass() && this.getName() == field.getName() && this.getType() == field.getType();
        }
        return false;
    }
    
    @Override
    public int hashCode() {
        return this.getDeclaringClass().getName().hashCode() ^ this.getName().hashCode();
    }
    
    @Override
    public String toString() {
        int modifiers = this.getModifiers();
        return ((modifiers == 0) ? "" : (Modifier.toString(modifiers) + " ")) + this.getType().getTypeName() + " " + this.getDeclaringClass().getTypeName() + "." + this.getName();
    }
    
    public String toGenericString() {
        int modifiers = this.getModifiers();
        return ((modifiers == 0) ? "" : (Modifier.toString(modifiers) + " ")) + this.getGenericType().getTypeName() + " " + this.getDeclaringClass().getTypeName() + "." + this.getName();
    }
    
    @CallerSensitive
    public Object get(Object o) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        return this.getFieldAccessor(o).get(o);
    }
    
    @CallerSensitive
    public boolean getBoolean(Object o) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        return this.getFieldAccessor(o).getBoolean(o);
    }
    
    @CallerSensitive
    public byte getByte(Object o) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        return this.getFieldAccessor(o).getByte(o);
    }
    
    @CallerSensitive
    public char getChar(Object o) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        return this.getFieldAccessor(o).getChar(o);
    }
    
    @CallerSensitive
    public short getShort(Object o) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        return this.getFieldAccessor(o).getShort(o);
    }
    
    @CallerSensitive
    public int getInt(Object o) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        return this.getFieldAccessor(o).getInt(o);
    }
    
    @CallerSensitive
    public long getLong(Object o) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        return this.getFieldAccessor(o).getLong(o);
    }
    
    @CallerSensitive
    public float getFloat(Object o) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        return this.getFieldAccessor(o).getFloat(o);
    }
    
    @CallerSensitive
    public double getDouble(Object o) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        return this.getFieldAccessor(o).getDouble(o);
    }
    
    @CallerSensitive
    public void set(Object o, Object o2) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        this.getFieldAccessor(o).set(o, o2);
    }
    
    @CallerSensitive
    public void setBoolean(Object o, boolean b) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        this.getFieldAccessor(o).setBoolean(o, b);
    }
    
    @CallerSensitive
    public void setByte(Object o, byte b) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        this.getFieldAccessor(o).setByte(o, b);
    }
    
    @CallerSensitive
    public void setChar(Object o, char c) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        this.getFieldAccessor(o).setChar(o, c);
    }
    
    @CallerSensitive
    public void setShort(Object o, short n) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        this.getFieldAccessor(o).setShort(o, n);
    }
    
    @CallerSensitive
    public void setInt(Object o, int n) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        this.getFieldAccessor(o).setInt(o, n);
    }
    
    @CallerSensitive
    public void setLong(Object o, long n) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        this.getFieldAccessor(o).setLong(o, n);
    }
    
    @CallerSensitive
    public void setFloat(Object o, float n) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        this.getFieldAccessor(o).setFloat(o, n);
    }
    
    @CallerSensitive
    public void setDouble(Object o, double n) throws IllegalArgumentException, IllegalAccessException {
        if (!this.override && !Reflection.quickCheckMemberAccess(this.clazz, this.modifiers)) {
            this.checkAccess(Reflection.getCallerClass(), this.clazz, o, this.modifiers);
        }
        this.getFieldAccessor(o).setDouble(o, n);
    }
    
    private FieldAccessor getFieldAccessor(Object o) throws IllegalAccessException {
        int override = this.override;
        FieldAccessor fieldAccessor = override ? this.overrideFieldAccessor : this.fieldAccessor;
        return (fieldAccessor != null) ? fieldAccessor : this.acquireFieldAccessor((boolean)override);
    }
    
    private FieldAccessor acquireFieldAccessor(boolean b) {
        FieldAccessor fieldAccessor = null;
        if (this.root != null) {
            fieldAccessor = this.root.getFieldAccessor(b);
        }
        if (fieldAccessor != null) {
            if (b) {
                this.overrideFieldAccessor = fieldAccessor;
            }
            else {
                this.fieldAccessor = fieldAccessor;
            }
        }
        else {
            fieldAccessor = Field.reflectionFactory.newFieldAccessor(this, b);
            this.setFieldAccessor(fieldAccessor, b);
        }
        return fieldAccessor;
    }
    
    private FieldAccessor getFieldAccessor(boolean b) {
        return b ? this.overrideFieldAccessor : this.fieldAccessor;
    }
    
    private void setFieldAccessor(FieldAccessor fieldAccessor, boolean b) {
        if (b) {
            this.overrideFieldAccessor = fieldAccessor;
        }
        else {
            this.fieldAccessor = fieldAccessor;
        }
        if (this.root != null) {
            this.root.setFieldAccessor(fieldAccessor, b);
        }
    }
    
    @Override
    public <T extends Annotation> T getAnnotation(Class<T> obj) {
        Objects.requireNonNull(obj);
        return (T)obj.cast(this.declaredAnnotations().get(obj));
    }
    
    @Override
    public <T extends Annotation> T[] getAnnotationsByType(Class<T> clazz) {
        Objects.requireNonNull(clazz);
        return AnnotationSupport.getDirectlyAndIndirectlyPresent(this.declaredAnnotations(), clazz);
    }
    
    @Override
    public Annotation[] getDeclaredAnnotations() {
        return AnnotationParser.toArray(this.declaredAnnotations());
    }
    
    private synchronized Map<Class<? extends Annotation>, Annotation> declaredAnnotations() {
        if (this.declaredAnnotations == null) {
            Field root = this.root;
            if (root != null) {
                this.declaredAnnotations = root.declaredAnnotations();
            }
            else {
                this.declaredAnnotations = AnnotationParser.parseAnnotations(this.annotations, SharedSecrets.getJavaLangAccess().getConstantPool(this.getDeclaringClass()), this.getDeclaringClass());
            }
        }
        return this.declaredAnnotations;
    }
    
    private native byte[] getTypeAnnotationBytes0();
    
    public AnnotatedType getAnnotatedType() {
        return TypeAnnotationParser.buildAnnotatedType(this.getTypeAnnotationBytes0(), SharedSecrets.getJavaLangAccess().getConstantPool(this.getDeclaringClass()), this, this.getDeclaringClass(), this.getGenericType(), TypeAnnotation.TypeAnnotationTarget.FIELD);
    }
}
